Object subclass: #Basic1Activity
	instanceVariableNames: 'address oself '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!


!Basic1Activity methodsFor: 'initialize'!

activeObject: anActiveObject
	oself := anActiveObject!

privateInitialize
	"Default initialize: do nothing."
	"It is not named initialize to ensure there is no name clash/shadowing with user (re)definition of initialize."! !

!Basic1Activity methodsFor: 'accessing'!

mailBox
	^address mailBox! !

!Basic1Activity methodsFor: 'address creation'!

computeAddressClass: activeObjectAddressClass
	"Compute and return the address class of the active object.
	Look for the most specialized between the address class specified/computed by the active object,
	and the one specified/computed by the activity itself."

	^(activeObjectAddressClass
		sub: self addressClass
		ifUnrelated: [self error: 'inconsistency between address classes'])!

createAndStartAddressWithAddressClass: activeObjectAddressClass
	"Create the address of the active object, activate it and return it."
	"Assume that method startActivity: does not specify a return value."

	^(self computeAddressClass: activeObjectAddressClass) new
		startActivity: self! !

!Basic1Activity methodsFor: 'activity setting'!

body
	"Define the behavior body."
	"Default body is an endless loop to serially consume successive pending messages."

	[true] whileTrue:
		[self acceptNextMessage]!

createProcess
	"Create and return the activity process which consumes the body."

	^[self body] newProcess!

start
	"Start the activity by starting the process which consumes the body."

	self createProcess resume!

startWithAddress: anAddress
	"Initialize and start the activity.
	After initializing the address, (and before starting the process),
	perform the generic initialization."

	address := anAddress.		"The activity has reference to the address."
	oself aself: address.			"The active object also will."
	self privateInitialize.
	self start						"Start the (process implementing the) activity."! !

!Basic1Activity methodsFor: 'message handling'!

acceptMessage: aMessage
	"Accept a message."
	"Default is to compute (perform) it."

	self performMessage: aMessage!

acceptNextMessage
	"Accepting next message."
	"(It gets suspended until there is one, see method nextMessage)."

	self acceptMessage: self nextMessage!

basicNextMessage
	"Remove next message from the mailbox."
	"Note that this fetching is suspended until the mailbox is not empty,
	thanks to the SharedQueue standard class."

	"This method is basic, that is must not be redefined."

	^self mailBox next!

nextMessage
	"Remove next message from the mailbox."
	"Note that this fetching is suspended until the mailbox is not empty,
	thanks to the SharedQueue standard class."

	^self mailBox next!

performMessage: aMessage
	"Perform (compute) a message."
	"Default is for the active object to actually perform it and return the value."

	^oself performMessage: aMessage! !

!Basic1Activity methodsFor: 'default classes'!

addressClass
	^Basic1Address! !

Basic1Activity subclass: #Basic2Activity
	instanceVariableNames: 'process '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic2Activity comment:
'Class Basic2ActorBehavior (paired with class Basic2Actor) offers a few more facilities than basic kernel classes Basic1Actor and Basic1ActorBehavior.

Class Basic2ActorBehavior provides a handle onto the background process which allow check, control (suspend, resume and terminate). The terminate action will be useful to do some cleanup in order to recover background processes, see ActorBehavior class methods classCleanUp and cleanUp.

It defines a new basic behavior process (block) in the process patterns library. This second one (the first and default process pattern accepts serially all incoming messages) accepts only one message. It will prove useful when implementing actors with multiple behaviors (see class ReplaceActorBehavior).

Class Basic2ActorBehavior also provides three generic event handling methods which are associated to actor events (receiving a message, accepting it, and completing its activation) which may be redefined for many purposes (trace, schedule, step...). These methods are respectively named eventReceive:, eventAccept:, and eventComplete:. They take the message on which event takes place as an argument. Note that these methods are user-oriented and are mostly used to trace events onto the transcript window (see class ActorBehavior class methods to automatically set up such tracing facilities). Kernel event methods also exist (and by default just dispatch onto user event methods). This is because generic events will be used to implement some synchronization mechanisms (like in class CountersSynchroActorBehavior). Therefore we prefer to avoid name clashes between user events and system events.

Last, it provides a checking when creating an actor (method actor) that the behavior is not shared with another actor. (This would be catastrophic because it would break the synchronization protection).

Class method allScriptSelectors returns all script selectors defined within a behavior class hierarchy. This method will be specially useful for synchronization behavior classes.

Basic kernel classes such as Basic2ActorBehavior are basic classes.
The entry point in the Actalk kernel is the set of the two kernel classes Actor and ActorBehavior.

Instance Variables:

	process		<Process>	is the process implementing the autonomy of the behavior.'!


!Basic2Activity methodsFor: 'accessing'!

activeObject
	^oself!

process
	^process! !

!Basic2Activity methodsFor: 'activity setting'!

createProcess
	"Provide a handle to the process."

	^process := super createProcess! !

!Basic2Activity methodsFor: 'process control'!

resume
	"Resume the activity process.
	Assume that the process has been suspended."

	process resume!

suspend
	"Suspend the activity process.
	Assume there is a process running."

	process suspend!

terminate
	"Terminate the activity process.
	Useful for cleanup (reclaiming resources)."

	process isNil
		ifFalse: [process terminateNoSignal]! !

!Basic2Activity methodsFor: 'message handling'!

acceptMessage: aMessage
	"Add the triggering of generic event methods for accepting and completing computation of a message."

	self kernelEventAccept: aMessage.	
	super acceptMessage: aMessage.
	self kernelEventComplete: aMessage! !

!Basic2Activity methodsFor: 'events'!

kernelEventAccept: aMessage
	"This method is called when the activity accept a message."
	"This is the basic kernel method which should not be subclassed by application classes."

	oself eventAccept: aMessage!

kernelEventComplete: aMessage
	"This method is called once the activity completed computation of a message."
	"This is the basic kernel method which should not be subclassed by application classes."

	oself eventComplete: aMessage!

kernelEventReceive: aMessage
	"This method is called after the address has received a new message."
	"This is the basic kernel method which should not be subclassed by application classes."

	oself eventReceive: aMessage! !

MinimalObject subclass: #Basic1Address
	instanceVariableNames: 'mailBox activity '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic1Address comment:
'Class Basic1Actor is the class which implements the external shell of an actor (seen from the outside, and refered to by other actors).
It acts mostly as a container for the mailbox which will contain incoming messages and the behavior which will process them. (See class Basic1ActorBehavior comment).
Its main purpose is also to implement asynchronous message passing.

Basic kernel classes such as Basic1ActorBehavior are basic classes.
The entry point in the Actalk kernel is the set of the two kernel classes Actor and ActorBehavior.

Instance Variables:

	mailBox		<MailBox>		contains incoming messages.
	behavior	<ActorBehavior>	is the actual behavior of the actor, that is the object which will actually consume and process incoming messages.

Redefined error handling method doesNotUnderstand: calls method asynchronousSend: to send an asynchronous message.
Method setBehavior: sets and initializes the behavior of an actor.
Method actalkPrivateInitialize (also provided by class Basic1ActorBehavior) is a generic initialize method. It is automatically called when creating instances by redefinition of class method new. It is not named initialize to ensure there is no name clash/shadowing with possible user (re)definition of initialize. It creates a new mailbox for the actor.

Acknowledgement:
The organization of Actalk kernel into a first basic kernel has been introduced by Loic Lescaudron within the development of the Actalk full environment (including extended MVC and generic time-slicing scheduler) implemented in Smalltalk-80 version 2.5.'!


!Basic1Address methodsFor: 'initialize'!

privateInitialize
	"Initialize the mailbox, by default a MailBox (SharedQueue)."
	"It is not named initialize to ensure there is no name clash/shadowing with user (re)definition of initialize."

	mailBox := MailBox new! !

!Basic1Address methodsFor: 'accessing'!

activity
	^activity!

mailBox
	^mailBox! !

!Basic1Address methodsFor: 'activity setting'!

startActivity: anActivity
	"Initialize reference to the activity and then start it."

	activity := anActivity.
	activity startWithAddress: self! !

!Basic1Address methodsFor: 'message passing'!

asynchronousSend: aMessage inMessageQueue: queue
	"Generic asynchronous message sending adds the message to the specified message queue."

	queue nextPut: aMessage!

doesNotUnderstand: aMessage
	"A message sent to an active object will trigger an error and be received."
	"Assumption of triggering an error (being unknown) is ensured by class MinimalObject."

	^self receiveMessage: aMessage!

receiveMessage: aMessage
	"Receive a message is by default through an asynchronous send into the mailBox."
	"Note that, asynchronousSend will return default return value, that is self,
	because an asynchronous and unidirectional send cannot return a specific reply value."

	^self asynchronousSend: aMessage inMessageQueue: mailBox! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Basic1Address class
	instanceVariableNames: ''!


!Basic1Address class methodsFor: 'instance creation'!

new
	^super new privateInitialize! !

Basic1Address subclass: #Basic2Address
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic2Address comment:
'Class Basic2Actor (paired with class Basic2ActorBehavior) implements a few more facilities than basic kernel class Basic1Actor (and Basic1ActorBehavior).
The pair provides access to the behavior process, this is useful for control and cleanup. It also offers generic events associated to actor events of receiving, accepting, and completing a message invocation. See class Basic2ActorBehavior comment.
Class Basic2Actor also provides direct acces by method asObject to the behavior (thus violating its encapsulation and synchronization, but may be useful in some situations, see method comment).'!


!Basic2Address methodsFor: 'accessing'!

asObject
	"Return the behavior of an actor."
	"This bypasses the actor shell and thus considers an actor as an almost standard Smalltalk object.
	This could be used to consult the contents of the behavior.
	This facility should be used extremely cautiously because it clashes with the protection/serialization of the actor.
	Consequently it should be used to consult (read) but not to change (write) the state of the behavior.
	Examples of use are when building user interfaces onto actors."

	^activity activeObject! !

!Basic2Address methodsFor: 'message passing'!

asynchronousSend: aMessage
	"Default asynchronous message send form."

	self asynchronousSend: aMessage inMessageQueue: mailBox!

asynchronousSend: aMessage inMessageQueue: queue
	"Add the triggering of generic event after receiving a message."

	"Note that the method kernelEventReceive: will also be defined in class UndefinedObject
	so that if there is no behavior yet (for instance in case of FutureActors) nil does not signal an error."

	super asynchronousSend: aMessage inMessageQueue: queue.
	activity kernelEventReceive: aMessage!

resend: aMessage
	"Resend the message by enqueueing it to the mailbox.
	Useful to implement (naive) delay of a message when the conditions of its acceptance are not yet fullfilled.
	(See for instance naive implementation of synchronization in class BoundedBuffer2, category Actalk-Ex-BoundedBuffer).

	WARNING: This may result in an infinite accept/resend cycle. (See below).

	Also notice that it does not call the method asynchronousSend:,
	this is because we don't want to trigger the receive event again."

	mailBox nextPut: aMessage.
	Processor yield		"Give way to other processes. This is NECESSARY.
						Otherwise there is strong chance this actor ALWAYS keep the processor.
						(By infinitely accepting and resending the same message)."!

resendWithoutYield: aMessage
	"Resend the message by enqueueing it to the mailbox.
	Useful to implement (naive) delay of a message when the conditions of its acceptance are not yet fullfilled.

	WARNING: This MAY result in an infinite accept/resend cycle.
	No processor yield is performed. (NO PROTECTION!!).
	However this is useful when a yield is unneeded in case of no accept loop
	(this is the case for subprocesses computing only one message).
	(See for instance naive implementation of synchronization in method doesNotUnderstand:
	of class EmptyBoundedBuffer within category Actalk-Ext-Replace-Examples.
	Also see class CountersActorBehavior within category Actalk-Synchro-Counters.)

	Also notice that it does not call the method asynchronousSend:,
	this is because we don't want to trigger the receive event again."

	mailBox nextPut: aMessage! !

!Basic2Address methodsFor: 'reset'!

reset
	"Full reset (mailbox and activity) and restart of an acyive object address."

	self privateInitialize.
	activity terminate.
	activity startWithAddress: self! !

!Basic2Address methodsFor: 'process control'!

resume
	"Resume the activity process."

	activity resume!

suspend
	"Suspend the activity process."

	activity suspend!

terminate
	"Terminate  the activity process.
	Useful for cleanup (reclaiming resources)."

	activity terminate! !

!Basic2Address methodsFor: 'printing'!

printOn: aStream
	"External representation of an actor: & followed by the behavior. Ex: &a Counter."

	aStream nextPut: $&.
	activity activeObject printOn: aStream! !

Object subclass: #Basic1ActiveObject
	instanceVariableNames: 'aself '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic1ActiveObject comment:
'Class Basic1ActorBehavior is the class which implements the behavior of an actor.
The behavior is the object which eventually consumes incoming messages and processes them.

Basic kernel classes such as Basic1ActorBehavior are basic classes.
The entry point in the Actalk kernel is the set of the two kernel classes Actor and ActorBehavior.

See more insights into how to define and use actors in class ActorBehavior comment.

Instance Variables:

	aself	<Actor>	is the actual actor owning this behavior.

aself is useful to send recursive messages (asynchronously) to the actor (through its mailbox).
Pseudo-variable self can be still used and will imply a standard recursive message passing (directly) to the behavior.

Method initializeAndSetAself: initializes the behavior.
Method actalkPrivateInitialize (also provided by class Basic1Actor) is a generic initialize method. Default does nothing.
Method setBehavior: sets its process by calling method behaviorProcessBlock which dispatches to the specified process behavior (default is accepting serially all pending messages).
Method acceptNextMessage decomposes into two generic methods: acceptMessage: and nextMessage (default is performing next message from the mailbox).
Method actorClass returns the default class for actors (here Basic1ActorBehavior).

Acknowledgements: Loic Lescaudron initially designed the basic kernel library (classes Basic1Actor(Behavior) and Basic2Actor(Behavior).'!


!Basic1ActiveObject methodsFor: 'initialize'!

aself: anAddress
	"Initialize the address (aself) of the active object.
	Then, perform the generic initialization."

	aself := anAddress.
	self privateInitialize!

privateInitialize
	"Default initialize: do nothing."
	"It is not named initialize to ensure there is no name clash/shadowing with user (re)definition of initialize."! !

!Basic1ActiveObject methodsFor: 'activity creation'!

active
	"Create an active object."
	"Default classes of the activity (and the address), e.g., classes Activity (and Address)
	are specified and returned by methods activityClass (and addressClass)."
	"This is similar to method defaultControllerClass of class View."

	^self activity: self activityClass address: self addressClass!

activity: activityClass address: addressClass
	"Create an active object whose classes of activity and address are specified."
	"Assume that method startWithAddressClass: returns the new address."

	^(activityClass new activeObject: self) createAndStartAddressWithAddressClass: addressClass! !

!Basic1ActiveObject methodsFor: 'default classes'!

activityClass
	^Basic1Activity!

addressClass
	^Basic1Address! !

Basic1ActiveObject subclass: #Basic2ActiveObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic2ActiveObject comment:
'Class Basic2ActorBehavior (paired with class Basic2Actor) offers a few more facilities than basic kernel classes Basic1Actor and Basic1ActorBehavior.

Class Basic2ActorBehavior provides a handle onto the background process which allow check, control (suspend, resume and terminate). The terminate action will be useful to do some cleanup in order to recover background processes, see ActorBehavior class methods classCleanUp and cleanUp.

It defines a new basic behavior process (block) in the process patterns library. This second one (the first and default process pattern accepts serially all incoming messages) accepts only one message. It will prove useful when implementing actors with multiple behaviors (see class ReplaceActorBehavior).

Class Basic2ActorBehavior also provides three generic event handling methods which are associated to actor events (receiving a message, accepting it, and completing its activation) which may be redefined for many purposes (trace, schedule, step...). These methods are respectively named eventReceive:, eventAccept:, and eventComplete:. They take the message on which event takes place as an argument. Note that these methods are user-oriented and are mostly used to trace events onto the transcript window (see class ActorBehavior class methods to automatically set up such tracing facilities). Kernel event methods also exist (and by default just dispatch onto user event methods). This is because generic events will be used to implement some synchronization mechanisms (like in class CountersSynchroActorBehavior). Therefore we prefer to avoid name clashes between user events and system events.

Last, it provides a checking when creating an actor (method actor) that the behavior is not shared with another actor. (This would be catastrophic because it would break the synchronization protection).

Class method allScriptSelectors returns all script selectors defined within a behavior class hierarchy. This method will be specially useful for synchronization behavior classes.

Basic kernel classes such as Basic2ActorBehavior are basic classes.
The entry point in the Actalk kernel is the set of the two kernel classes Actor and ActorBehavior.

Instance Variables:

	process		<Process>	is the process implementing the autonomy of the behavior.'!


!Basic2ActiveObject methodsFor: 'accessing'!

activity
	^aself activity! !

!Basic2ActiveObject methodsFor: 'activity creation'!

active
	"Check if the active object has already been used (activated)."

	^aself isNil
		ifTrue: [super active]
		ifFalse: [self error: 'has already been activated']! !

!Basic2ActiveObject methodsFor: 'process control'!

suspend
	"Suspend the activity process."

	aself suspend!

terminate
	"Terminate the activity process.
	Useful for cleanup (reclaiming resources)."

	aself terminate! !

!Basic2ActiveObject methodsFor: 'events'!

eventAccept: aMessage
	"This method is called when the active object accepts a message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' accept (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventAccept with: aMessage.
step:
	self halt.
	"!

eventComplete: aMessage
	"This method is called once the active object completed computation of a message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' complete (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventComplete with: aMessage.
step:
	self halt.
	"!

eventReceive: aMessage
	"This method is called once the active object has received a new message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' receive (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventReceive with: aMessage.
step:
	self halt.
	"! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Basic2ActiveObject class
	instanceVariableNames: ''!


!Basic2ActiveObject class methodsFor: 'accessing method dictionary'!

allScriptSelectors
	"Answer the set of all the script selectors known by a behavior class."

	^self allSelectorsInCategory: #script untilClass: ActiveObject! !

SharedQueue subclass: #MailBox
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
MailBox comment:
'Class MailBox is a subclass of standard class SharedQueue.
SharedQueue provides basic functionalities of a mailbox to receive and retrieve messages (plus mutual exclusion and suspension while the mailbox is empty).
Class MailBox implements some accessing methods specific to Actalk (looking for specific patterns and incrementally look for a message). See the protocol message accesing. Note that these are in someway low-level implementation routines.'!


!MailBox methodsFor: 'message accessing'!

at: i
	"Look for and fetch the ith message in the mailbox. No check/suspension."

	^[accessProtect critical: [contents at: i]] valueUninterruptably!

atIndex: i
	"Look for and fetch the ith message in the mailbox.
	Suspend until it exists if not yet (by waiting for new messages as many times as needed)."
	"Useful for class GuardsSynchroActorBehavior."
	"WARNING: this method is safe if only one behavior at a time uses it.
	It is unsafe for concurrent use by many behaviors
	(because it makes assumption on the monotony of the mailbox)."
	"The manipulation of readSynch semaphore is somewhat tricky.
	We need to because we want to suspend until a new message comes
	BUT we don't want to decrement the number of items in the queue
	(until we actually find and remove a matching message)."
	"Used by class GuardsSynchroActorBehavior."

	| numberOfWaits |		"We need to record number of waits on readSynch semaphore"
	numberOfWaits := 0.		"in order to restore them afterwards."
		"Because this number represents the number of messages in the mailbox."
	[[accessProtect critical: [i > contents size]] valueUninterruptably]
		whileTrue:
			[readSynch wait.		"Wait for next incoming message."
			numberOfWaits := numberOfWaits + 1].
	numberOfWaits timesRepeat: [readSynch signal].	"Restore the semaphore excessive signals."
	^contents at: i!

checkFirstMessageAmong: collectionOfSelectors
	"Look for the first message matching one of the selectors.
	As opposed to method firstMessageAmong: which is blocking,
	unBlockingLookForFirstMessageAmong: is not.
	If a matching message is found it removes and returns it. Otherwise it returns nil."
	"May be useful for class AcceptActorBehavior."

	^self internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: true!

contents
	"Answer the contents (ordered collection) of the mailbox."

	^[accessProtect critical: [contents]] valueUninterruptably!

firstMessageAmong: collectionOfSelectors
	"Look for the first message matching one of the selectors.
	It is blocking and suspend until it finds any matching one.
	(This means it suspends onto incoming messages as long as necessary)."
	"The manipulation of readSynch semaphore excessive signals is somewhat tricky.
	We need to because we want to suspend until a new message comes
	BUT we don't want to decrement the number of items in the queue
	(until we actually find and remove a matching message)."
	"Used by classes AbclActorBehavior, AcceptActorBehavior and StateSynchroActorBehavior."

	| message numberOfWaits |
		"First instant look for a matching message."
	message := self internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: true.
		"We need to record number of waits on readSynch semaphore in order to restore them afterwards."
		"Because this number represents the number of messages in the mailbox."
	numberOfWaits := 0.	
	[message isNil]		"While no matching message has been found yet,"
		whileTrue:		"we need to check again repeatedly after each new incoming message."
			[readSynch wait.		"Wait for next incoming message."
			numberOfWaits := numberOfWaits + 1.
				"The flag/argument of withReadSynchWait: is false, because we already waited for current message."
			message := self internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: false].
	numberOfWaits - 1 timesRepeat: [readSynch signal].	"Restore the semaphore excessive signals."
	^message										"(minus one, because one for the removed item)."!

internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: isWait
	"Look for the first message matching one of the selectors.
	It is not blocking and returns nil if no matching message is found."
	"The isWait flag tells if a wait onto the readSynch semaphore should be produced
	in case a matching message has been found and will be removed.
	Producing a wait is the usual case to tell that one item has been removed.
	However in some case (call from the loop of method firstMessageAmong:)
	the wait has already been produced, and doing it again would suspend computation."

	| message |
	[accessProtect critical:
		[message := contents
					detect: [:msg | collectionOfSelectors includes: msg selector]
					ifNone: [].
		message isNil				"If found, we remove (and then return) it."
			ifFalse:
				[isWait ifTrue: [readSynch wait].	"We know we don't have to wait, but we want to tell we remove one item."
				contents remove: message]]] valueUninterruptably.
	^message!

removeAtIndex: i
	"Remove and return the ith message of the mailbox."
	"Useful for class GuardsSynchroActorBehavior."
	"WARNING: this method is safe if only one behavior uses it.
	It is unsafe for concurrent use by many behaviors
	(because it makes assumption on the monotony of the mailbox)."

	readSynch wait.		"We assume that we don't have to wait, but we want to tell we remove one item."
	^[accessProtect critical: [contents removeAtIndex: i]] valueUninterruptably! !

!MailBox methodsFor: 'printing'!

printOn: aStream
	"Print as its contents (an OrderedCollection)."

	contents printOn: aStream! !
