Address subclass: #Actor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Compatibility'!


Activity subclass: #ActorBehavior
	instanceVariableNames: 'aself '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Compatibility'!


!ActorBehavior methodsFor: 'initialize'!

aself: anAddress
	aself := anAddress! !

!ActorBehavior methodsFor: 'address creation'!

computeAddressClass: addressClass
	^addressClass! !

!ActorBehavior methodsFor: 'message handling'!

performMessage: aMessage
	"The activity directly perform (compute) the message."

	^self basicPerformMessage: aMessage! !

!ActorBehavior methodsFor: 'actor creation'!

actor
	"Create an actor, once the behavior is specified."
	"The class of the actor (e.g., class Actor) is specified and returned by method actorClass."
	"This is similar to method defaultControllerClass of class View."

	^self activeObject: self;
		createAndStartAddressWithAddressClass: self actorClass!

actorClass
	"Specify the associated class of actors, by default class Actor."
	"Is redefined when defining new pairs of classes *ActorBehavior and *Actor."

	^Actor! !

!ActorBehavior methodsFor: 'events'!

eventAccept: aMessage
	"This method is called when the active object accepts a message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' accept (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventAccept with: aMessage.
step:
	self halt.
	"!

eventComplete: aMessage
	"This method is called once the active object completed computation of a message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' complete (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventComplete with: aMessage.
step:
	self halt.
	"!

eventReceive: aMessage
	"This method is called once the active object has received a new message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' receive (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventReceive with: aMessage.
step:
	self halt.
	"!

kernelEventAccept: aMessage
	"This method is called when the activity accept a message."
	"This is the basic kernel method which should not be subclassed by application classes."

	self eventAccept: aMessage!

kernelEventComplete: aMessage
	"This method is called once the activity completed computation of a message."
	"This is the basic kernel method which should not be subclassed by application classes."

	self eventComplete: aMessage!

kernelEventReceive: aMessage
	"This method is called after the address has received a new message."
	"This is the basic kernel method which should not be subclassed by application classes."

	self eventReceive: aMessage! !

!ActorBehavior methodsFor: 'replying'!

reply: value withSelector: replySelector
	"Reply with a specific reply selector which needs to be computed.
	Basically it has been passed as an argument. See example in class Counter."

	self perform: replySelector with: value! !

!ActorBehavior methodsFor: 'continuation creation'!

continuationBlock: aBlock
	"Create a block continuation actor with the block (closure) as a behavior.
	The block has two arguments: the replied value, and the actor continuation actor (usually named me)."
	"See class BlockContinuation."

	^(BlockContinuation new behaviorBlock: aBlock) active!

singleReplyContinuationBlock: aBlock
	"Create a block continuation actor with the block (closure) as a behavior.
	The block has two arguments: the replied value, and the actor continuation actor (usually named me)."
	"See class SingleReplyBlockContinuation."

	^(SingleReplyBlockContinuation new behaviorBlock: aBlock) active! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ActorBehavior class
	instanceVariableNames: ''!


!ActorBehavior class methodsFor: 'accessing method dictionary'!

allScriptSelectors
	"Answer the set of all the script selectors known by a behavior class."

	^self allSelectorsInCategory: #script untilClass: ActorBehavior! !

ActorBehavior subclass: #ABCounter
	instanceVariableNames: 'contents '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Compatibility'!
ABCounter comment:
'Class Counter is the seminal (and most simple) example of object, in this case actor.
It shows the three main differences with standard Smalltalk programming:

	Class Counter is defined as a subclass of class ActorBehavior and not of class Object.

	The actor is created by sending the message actor to the newly created behavior.

	Consulting the contents implies passing the reply destination (e.g., the predefined printer Print) because in standard Actalk asynchronous message passing there is no implicit reply. This also means that standard Smalltalk-80 caret (^) to return the value is of no use (except for escaping the method). Replying must be explicitly stated as a message sending (with reply: selector) to the reply destination, either passed as an argument or stored (for example by a continuation, see category Actalk-Kernel-JoinContinuations).

Note that all examples (should) have (at least) one example method, defined as a class method within protocol example (thus following standard Smalltalk-80 convention).

Instance Variables:

	contents		<Integer>	the contents of the counter.

In order to have homogeneous naming for methods passing reply destinations, we use the following conventions (in addition to using reply: as the convention for reply selector):
	add AndReplyTo: to the end of a selector when there is no other argument (besides the reply destination),
		example is selector contentsAndReplyTo:
	add replyTo: to the end of a selector when there are other arguments,
		example is selector n:replyTo: of class Factorial.
In case of passing the reply selector this becomes AndReplyTo:withSelector: or replyTo:withSelector:, example is selector contentsAndReplyTo:withSelector:.'!


!ABCounter methodsFor: 'initialize'!

contents: anInteger
	contents := anInteger! !

!ABCounter methodsFor: 'script'!

consultAndReplyTo: r
	"As for asynchronous message passing, returning reply needs to be made explicit.
	The reply destination (r) is passed as an argument."

	r reply: contents!

consultAndReplyTo: r withSelector: replySelector
	"If the reply selector is variable (e.g., reply: or store: or print:...),
	we need to pass it as an argument (replySelector) and compute it
	(method reply:withSelector: of class ActorBehavior)."

	r reply: contents withSelector: replySelector!

incr
	contents := contents + 1!

reset
	contents := 0! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ABCounter class
	instanceVariableNames: ''!


!ABCounter class methodsFor: 'example'!

example
	"self example"

	(self new contents: 100) actor
				incr;
				incr;
				consultAndReplyTo: Print;
				incr;
				consultAndReplyTo: Print! !
